from __future__ import annotations

import logging

from pygama import lgdo
from pygama.raw.compass.compass_config_parser import compass_config_to_struct
from pygama.raw.data_decoder import DataDecoder
from pygama.raw.raw_buffer import RawBuffer

log = logging.getLogger(__name__)


class CompassHeaderDecoder(DataDecoder):
    """
    Decode CoMPASS header data. Also, read in CoMPASS config data if provided using the compass_config_parser
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.config = None  # initialize to none, because compass_config_to_struct always returns a struct

    def decode_header(
        self, in_stream: bytes, config_file: str = None, wf_len: int = None
    ) -> dict:
        """Decode the CoMPASS file header, and add CoMPASS config data to the header, if present.

        Parameters
        ----------
        in_stream
            The stream of data to have its header decoded
        config_file
            The config file for the CoMPASS data, if present
        wf_len
            The length of the first waveform in the file, only pre-calculated when the config_file is none

        Returns
        -------
        config
            A dict containing the header information, as well as the important config information
            of wf_len and num_enabled_channels
        """
        self.config = compass_config_to_struct(config_file, wf_len)

        config_names = [
            "energy_channels",  # energy is given in channels (0: false, 1: true)
            "energy_calibrated",  # energy is given in keV/MeV, according to the calibration (0: false, 1: true)
            "energy_short",  # energy short is present (0: false, 1: true)
            "waveform_samples",  # waveform samples are present (0: false, 1: true)
        ]
        header_in_bytes = in_stream.read(
            2
        )  # we always have to read in the first 2 bytes of the header for CoMPASS v2 files
        header_in_binary = bin(int.from_bytes(header_in_bytes, byteorder="little"))
        header_as_list = str(header_in_binary)[
            ::-1
        ]  # reverse it as we care about bit 0, bit 1, etc.

        for i, name in enumerate(config_names):
            if name in self.config:
                log.warning(f"{name} already in self.config. skipping...")
                continue
            value = int(header_as_list[i])
            self.config.add_field(
                str(name), lgdo.Scalar(value)
            )  # self.config is a struct

        return self.config

    def make_lgdo(self, key: int = None, size: int = None) -> lgdo.Struct:
        if self.config is None:
            raise RuntimeError(
                "self.config still None, need to decode header before calling make_lgdo"
            )
        return self.config  # self.config is already an lgdo, namely it is a struct

    def buffer_is_full(self, rb: RawBuffer) -> bool:
        return rb.loc > 0



import numpy as np
import pytest

from pygama.dsp.errors import DSPFatal
from pygama.dsp.processors import discrete_wavelet_transform


def test_discrete_wavelet_transform(compare_numba_vs_python):
    """Testing function for the discrete_wavelet_transform processor."""

    # set up values to use for each test case
    len_wf_in = 16
    wave_type = "haar"
    level = 2
    len_wf_out = 4

    # ensure the DSPFatal is raised for a negative level
    with pytest.raises(DSPFatal):
        discrete_wavelet_transform(wave_type, -1)

    # ensure that a valid input gives the expected output
    w_in = np.ones(len_wf_in)
    w_out = np.empty(len_wf_out)
    w_out_expected = np.ones(len_wf_out) * 2 ** (level / 2)

    dwt_func = discrete_wavelet_transform(wave_type, level)
    assert np.allclose(
        compare_numba_vs_python(dwt_func, w_in, w_out),
        w_out_expected,
    )

    # ensure that if there is a nan in w_in, all nans are outputted
    w_in = np.ones(len_wf_in)
    w_in[4] = np.nan
    w_out = np.empty(len_wf_out)

    dwt_func = discrete_wavelet_transform(wave_type, level)
    assert np.all(np.isnan(compare_numba_vs_python(dwt_func, w_in, w_out)))



from pathlib import Path

from pygama.vis import WaveformBrowser

config_dir = Path(__file__).parent / "configs"


def test_basics(lgnd_test_data):
    wb = WaveformBrowser(
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        "/geds/raw",
        dsp_config=f"{config_dir}/hpge-dsp-config.json",
        lines=["wf_blsub", "wf_trap", "trapEmax"],
        legend=["waveform", "trapezoidal", "energy = {trapEmax:0.1f}"],
        styles="seaborn-v0.8",
        n_drawn=2,
        x_lim=("20*us", "60*us"),
        x_unit="us",
    )

    wb.draw_next()
    wb.draw_entry(24)
    wb.draw_entry((2, 24))


def test_entry_mask(lgnd_test_data):
    wb = WaveformBrowser(
        lgnd_test_data.get_path("lh5/LDQTA_r117_20200110T105115Z_cal_geds_raw.lh5"),
        "/geds/raw",
        entry_list=[
            7,
            9,
            25,
            27,
            33,
            38,
            46,
            52,
            57,
            59,
            67,
            71,
            72,
            82,
            90,
            92,
            93,
            94,
            97,
        ],
    )

    wb.draw_next()



from __future__ import annotations

from typing import Callable

import numpy as np
from iminuit import Minuit
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs

from .pole_zero import double_pole_zero, pole_zero


class Model:
    """A class representing the function to minimize."""

    errordef = Minuit.LEAST_SQUARES

    # the constructor
    def __init__(
        self, func: Callable, w_in: np.ndarray, baseline: float, beg: int, end: int
    ) -> None:
        self.func = func
        self.x = np.arange(beg, end)
        self.y = np.asarray(w_in, dtype=np.float64) - baseline
        self.beg = beg
        self.end = end

    # the function to minimize
    def __call__(self, args) -> np.ndarray:
        y_pz = self.func(self.y, *args)[self.beg : self.end]
        return np.abs(
            np.sum(self.x) * np.sum(y_pz) - len(self.x) * np.sum(self.x * y_pz)
        )


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32[:])",
        "void(float64[:], float64, float64, float64, float64, float64[:])",
    ],
    "(n),(),(),(),()->()",
    **nb_kwargs,
    forceobj=True,
)
def optimize_1pz(
    w_in: np.ndarray,
    a_baseline_in: float,
    t_beg_in: int,
    t_end_in: int,
    p0_in: float,
    val0_out: float,
) -> None:
    """Find the optimal, single pole-zero cancellation's parameter
    by minimizing the slope in the waveform's specified time range.

    Parameters
    ----------
    w_in
        the input waveform.
    a_baseline_in
        the resting baseline.
    t_beg_in
        the lower bound's index for the time range over
        which to optimize the pole-zero cancellation.
    t_end_in
        the upper bound's index for the time range over
        which to optimize the pole-zero cancellation.
    p0_in
        the initial guess of the optimal time constant.
    val0_out
        the output value of the best-fit time constant.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tau0": {
            "function": "optimize_1pz",
            "module": "pygama.dsp.processors",
            "args": ["waveform", "baseline", "0", "20*us", "500*us", "tau0"],
            "unit": "us"
        }

    See Also
    --------
    ~.pole_zero.pole_zero, .double_pole_zero
    """
    val0_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_baseline_in)
        or np.isnan(t_beg_in)
        or np.isnan(t_end_in)
        or np.isnan(p0_in)
    ):
        return

    if not np.floor(t_beg_in) == t_beg_in or not np.floor(t_end_in) == t_end_in:
        raise DSPFatal("The waveform index is not an integer")

    if (
        int(t_beg_in) < 0
        or int(t_beg_in) > len(w_in)
        or int(t_end_in) < 0
        or int(t_end_in) > len(w_in)
    ):
        raise DSPFatal("The waveform index is out of range")

    m = Minuit(
        Model(pole_zero, w_in, a_baseline_in, int(t_beg_in), int(t_end_in)), [p0_in]
    )
    m.print_level = -1
    m.strategy = 1
    m.errordef = Minuit.LEAST_SQUARES
    m.migrad()
    val0_out[0] = m.values[0]


@guvectorize(
    [
        "void(float32[:], float32, float32, float32, float32, float32, float32, float32[:], float32[:], float32[:])",
        "void(float64[:], float64, float64, float64, float64, float64, float64, float64[:], float64[:], float64[:])",
    ],
    "(n),(),(),(),(),(),()->(),(),()",
    **nb_kwargs,
    forceobj=True,
)
def optimize_2pz(
    w_in: np.ndarray,
    a_baseline_in: float,
    t_beg_in: int,
    t_end_in: int,
    p0_in: float,
    p1_in: float,
    p2_in: float,
    val0_out: float,
    val1_out: float,
    val2_out: float,
) -> None:
    """Find the optimal, double pole-zero cancellation's parameters by
    minimizing the slope in the waveform's specified time range.

    Parameters
    ----------
    w_in
        the input waveform.
    a_baseline_in
        the resting baseline.
    t_beg_in
        the lower bound's index for the time range over
        which to optimize the pole-zero cancellation.
    t_end_in
        the upper bound's index for the time range over
        which to optimize the pole-zero cancellation.
    p0_in
        the initial guess of the optimal, longer time constant.
    p1_in
        the initial guess of the optimal, shorter time constant.
    p2_in
        the initial guess of the optimal fraction.
    val0_out
        the output value of the best-fit, longer time constant.
    val1_out
        the output value of the best-fit, shorter time constant.
    val2_out
        the output value of the best-fit fraction.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "tau1, tau2, frac": {
            "function": "optimize_2pz",
            "module": "pygama.dsp.processors",
            "args": [
                "waveform", "baseline", "0", "20*us", "500*us",
                "20*us", "0.02", "tau1", "tau2", "frac"
            ],
            "unit": "us"
        }
    """
    val0_out[0] = np.nan
    val1_out[0] = np.nan
    val2_out[0] = np.nan

    if (
        np.isnan(w_in).any()
        or np.isnan(a_baseline_in)
        or np.isnan(t_beg_in)
        or np.isnan(t_end_in)
        or np.isnan(p0_in)
        or np.isnan(p1_in)
        or np.isnan(p2_in)
    ):
        return

    if not np.floor(t_beg_in) == t_beg_in or not np.floor(t_end_in) == t_end_in:
        raise DSPFatal("The waveform index is not an integer")

    if (
        int(t_beg_in) < 0
        or int(t_beg_in) > len(w_in)
        or int(t_end_in) < 0
        or int(t_end_in) > len(w_in)
    ):
        raise DSPFatal("The waveform index is out of range")

    m = Minuit(
        Model(double_pole_zero, w_in, a_baseline_in, int(t_beg_in), int(t_end_in)),
        [p0_in, p1_in, p2_in],
    )
    m.print_level = -1
    m.strategy = 1
    m.errordef = Minuit.LEAST_SQUARES
    m.migrad()
    val0_out[0] = m.values[0]
    val1_out[0] = m.values[1]
    val2_out[0] = m.values[2]



r"""
Utilities to manage data buffering for raw data conversion. This module manages
LGDO buffers and their corresponding output streams. Allows for one-to-many
mapping of input streams to output streams.

Primary Classes
---------------
:class:`.RawBuffer`: an LGDO (e.g. a table) along with buffer metadata, such as the
current write location, the list of keys (e.g. channels) that write to it, the
output stream it is associated with (if any), etc. Each
:class:`~.raw.data_decoder.DataDecoder` is associated with a
:class:`.RawBuffer` of a particular format.

:class:`.RawBufferList`: a collection of :class:`RawBuffer` with LGDO's that
all have the same structure (same type, same fields, etc., but the fields can
have different shape). A :class:`~.raw.data_decoder.DataDecoder` will write its
output to a :class:`.RawBufferList`.

:class:`.RawBufferLibrary`: a dictionary of :class:`RawBufferList`\ s, e.g. one
for each :class:`~.raw.data_decoder.DataDecoder`. Keyed by the decoder name.

:class:`.RawBuffer` supports a JSON short-hand notation, see
:meth:`.RawBufferLibrary.set_from_json_dict` for full specification.

Example JSON yielding a valid :class:`.RawBufferLibrary` is below. In the
example, the user would call ``RawBufferLibrary.set_from_json_dict(json_dict,
kw_dict)`` with ``kw_dict`` containing an entry for ``'file_key'``. The other
keywords ``{key}`` and ``{name}`` are understood by and filled in during
:meth:`.RawBufferLibrary.set_from_json_dict` unless overloaded in ``kw_dict``.
Note the use of the wildcard ``*``: this will match all other decoder names /
keys.

.. code-block :: json

    {
      "FCEventDecoder" : {
        "g{key:0>3d}" : {
          "key_list" : [ [24,64] ],
          "out_stream" : "$DATADIR/{file_key}_geds.lh5:/geds",
          "proc_spec": {
            "window": {"waveform", 10, 100, "windowed_waveform"}
          }
        },
        "spms" : {
          "key_list" : [ [6,23] ],
          "out_stream" : "$DATADIR/{file_key}_spms.lh5:/spms"
        },
        "puls" : {
          "key_list" : [ 0 ],
          "out_stream" : "$DATADIR/{file_key}_auxs.lh5:/auxs"
        },
        "muvt" : {
          "key_list" : [ 1, 5 ],
          "out_stream" : "$DATADIR/{file_key}_auxs.lh5:/auxs"
        }
      },
      "*" : {
        "{name}" : {
          "key_list" : [ "*" ],
          "out_stream" : "$DATADIR/{file_key}_{name}.lh5"
        }
      }
    }
"""
from __future__ import annotations

import os
from typing import Union

from pygama import lgdo
from pygama.lgdo.lh5_store import LH5Store
from pygama.raw.buffer_processor.buffer_processor import buffer_processor

LGDO = Union[lgdo.Scalar, lgdo.Struct, lgdo.Array, lgdo.VectorOfVectors]


class RawBuffer:
    r"""Base class to represent a buffer of raw data.

    A :class:`RawBuffer` is in essence a an LGDO object (typically a
    :class:`~.lgdo.table.Table`) to which decoded data will be written, along
    with some meta-data distinguishing what data goes into it, and where the
    LGDO gets written out. Also holds on to the current location in the buffer
    for writing.

    Attributes
    ----------
    lgdo
        the LGDO used as the actual buffer. Typically a
        :class:`~.lgdo.table.Table`. Set to ``None`` upon creation so that the
        user or a decoder can initialize it later.
    key_list
        a list of keys (e.g. channel numbers) identifying data to be written
        into this buffer. The key scheme is specific to the decoder with which
        the :class:`.RawBuffer` is associated. This is called `key_list`
        instead of `keys` to avoid confusion with the dict function
        :meth:`dict.keys`, i.e.  ``raw_buffer.lgdo.keys()``.
    out_stream
        the output stream to which the :class:`.RawBuffer`\ 's LGDO should be
        sent or written. A colon (``:``) can be used to separate the stream
        name/address from an in-stream path/port:
        - file example: ``/path/filename.lh5:/group``
        - socket example: ``198.0.0.100:8000``
    out_name
        the name or identifier of the object in the output stream.
    proc_spec
        a dictionary containing the following:
        - a DSP config file, passed as a dictionary, or as a path to a JSON file
        - an array containing: the name of an :class:`~.lgdo` object stored in the :class:`.RawBuffer` to be sliced,
        the start and end indices of the slice, and the new name for the sliced object
        - a dictionary of fields to drop
        - a dictionary of new fields and their return datatype
        these specifications are used to process the data with :meth:`.buffer_processor.buffer_processor.buffer_processor`,
        refer to the documentation for more details on how the format of `proc_spec` and how the processing is performed.
    """

    def __init__(
        self,
        lgdo: LGDO = None,
        key_list: list[int | str] = None,
        out_stream: str = "",
        out_name: str = "",
        proc_spec: dict = None,
    ) -> None:
        self.lgdo = lgdo
        self.key_list = [] if key_list is None else key_list
        self.out_stream = out_stream
        self.out_name = out_name
        self.proc_spec = proc_spec
        self.loc = 0
        self.fill_safety = 1

    def __len__(self) -> int:
        if self.lgdo is None:
            return 0
        if not hasattr(self.lgdo, "__len__"):
            return 1
        return len(self.lgdo)

    def is_full(self) -> bool:
        return (len(self) - self.loc) < self.fill_safety

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return (
            "RawBuffer(lgdo="
            + repr(self.lgdo)
            + ", key_list="
            + repr(self.key_list)
            + ", out_stream="
            + repr(self.out_stream)
            + ", out_name="
            + repr(self.out_name)
            + ", loc="
            + repr(self.loc)
            + ", fill_safety="
            + repr(self.fill_safety)
            + ", proc_spec="
            + repr(self.proc_spec)
            + ")"
        )


class RawBufferList(list):
    r"""A :class:`.RawBufferList` holds a collection of :class:`.RawBuffer`\ s
    of identical structure (same format LGDO's with the same fields).
    """

    def __init__(self) -> None:
        self.keyed_dict = None

    def get_keyed_dict(self) -> dict[int | str, RawBuffer]:
        r"""Returns a dictionary of :class:`.RawBuffer`\ s built from the
        buffers' `key_lists`.

        Different keys may point to the same buffer. Requires the buffers in
        the :class:`.RawBufferList` to have non-overlapping key lists.
        """
        if self.keyed_dict is None:
            self.keyed_dict = {}
            for rb in self:
                for key in rb.key_list:
                    self.keyed_dict[key] = rb
        return self.keyed_dict

    def set_from_json_dict(
        self, json_dict: dict, kw_dict: dict[str, str] = None
    ) -> None:
        """Set up a :class:`.RawBufferList` from a dictionary written in JSON
        shorthand. See :meth:`.RawBufferLibrary.set_from_json_dict` for details.

        Notes
        -----
        `json_dict` is changed by this function.
        """
        expand_rblist_json_dict(json_dict, {} if kw_dict is None else kw_dict)
        for name in json_dict:
            rb = RawBuffer()
            if "key_list" in json_dict[name]:
                rb.key_list = json_dict[name]["key_list"]
            if "out_stream" in json_dict[name]:
                rb.out_stream = json_dict[name]["out_stream"]
            if "proc_spec" in json_dict[name]:
                rb.proc_spec = json_dict[name][
                    "proc_spec"
                ]  # If you swap this with the next line, then key expansion doesn't work
            if "out_name" in json_dict[name]:
                rb.out_name = json_dict[name]["out_name"]
            else:
                rb.out_name = name
            self.append(rb)

    def get_list_of(self, attribute: str) -> list:
        """Return a list of values of :class:`.RawBuffer` attributes.

        Parameters
        ----------
        attribute
            The :class:`.RawBuffer` attribute queried to make the list.

        Returns
        -------
        values
            The list of values of `RawBuffer.attribute`.

        Examples
        --------
        >>> output_file_list = rbl.get_list_of('out_stream')
        """
        values = []
        for rb in self:
            if not hasattr(rb, attribute):
                continue
            val = getattr(rb, attribute)
            if val not in values:
                values.append(val)
        return values

    def clear_full(self) -> None:
        for rb in self:
            if rb.is_full():
                rb.loc = 0


class RawBufferLibrary(dict):
    r"""A :class:`.RawBufferLibrary` is a collection of
    :class:`.RawBufferList`\ s associated with the names of decoders that can
    write to them.
    """

    def __init__(self, json_dict: dict = None, kw_dict: dict[str, str] = None) -> None:
        if json_dict is not None:
            self.set_from_json_dict(json_dict, kw_dict)

    def set_from_json_dict(
        self, json_dict: dict, kw_dict: dict[str, str] = None
    ) -> None:
        r"""Set up a :class:`.RawBufferLibrary` from a dictionary written in
        JSON shorthand.

        Basic structure:

        .. code-block :: js

            {
            "list_name" : {
              "name" : {
                  "key_list" : [ "key1", "key2", "..." ],
                  "out_stream" : "out_stream_str",
                  "out_name" : "out_name_str" // (optional)
                  "proc_spec" : { // (optional)
                    "windowed":
                        ["waveform", 10, 100, "windowed_waveform"],

                  }
              }
            }

        By default ``name`` is used for the :class:`RawBuffer`\ 's ``out_name``
        attribute, but this can be overridden if desired by providing an
        explicit ``out_name``.

        Allowed shorthands, in order of expansion:

        * ``key_list`` may have entries that are 2-integer lists corresponding
          to the first and last integer keys in a contiguous range (e.g. of
          channels) that get stored to the same buffer. These simply get
          replaced with the explicit list of integers in the range. We use
          lists not tuples for JSON compliance.
        * The ``name`` can include ``{key:xxx}`` format specifiers, indicating
          that each key in ``key_list`` should be given its own buffer with the
          corresponding name.  The same specifier can appear in ``out_path`` to
          write the key's data to its own output path.
        * You may also include keywords in your ``out_stream`` and ``out_name``
          specification whose values get sent in via `kw_dict`. These get
          evaluated simultaneously with the ``{key:xxx}`` specifiers.
        * Environment variables can also be used in ``out_stream``. They get
          expanded after `kw_dict` is handled and thus can be used inside
          `kw_dict`.
        * ``list_name`` can use the wildcard ``*`` to match any other
          ``list_name`` known to a streamer.
        * ``out_stream`` and ``out_name`` can also include ``{name}``, to be
          replaced with the buffer's ``name``. In the case of
          ``list_name="*"``, ``{name}`` evaluates to ``list_name``.

        Parameters
        ----------
        json_dict
            loaded from a JSON file written in the allowed shorthand.
            `json_dict` is changed by this function.
        kw_dict
            dictionary of keyword-value pairs for substitutions into the
            ``out_stream`` and ``out_name`` fields.
        """
        for list_name in json_dict:
            if list_name not in self:
                self[list_name] = RawBufferList()
            self[list_name].set_from_json_dict(
                json_dict[list_name], {} if kw_dict is None else kw_dict
            )

    def get_list_of(self, attribute: str, unique: bool = True) -> list:
        """Return a list of values of :class:`.RawBuffer` attributes.

        Parameters
        ----------
        attribute
            The :class:`.RawBuffer` attribute queried to make the list.
        unique
            whether to remove duplicates.

        Returns
        -------
        values
            The list of values of `RawBuffer.attribute`.

        Examples
        --------
        >>> output_file_list = rbl.get_list_of('out_stream')
        """
        values = []
        for rb_list in self.values():
            values += rb_list.get_list_of(attribute)
        if unique:
            values = list(set(values))
        return values

    def clear_full(self) -> None:
        for rb_list in self.values():
            rb_list.clear_full()


def expand_rblist_json_dict(json_dict: dict, kw_dict: dict[str, str]) -> None:
    """Expand shorthands in a JSON dictionary representing a
    :class:`.RawBufferList`.

    See :meth:`.RawBufferLibrary.set_from_json_dict` for details.

    Notes
    -----
    The input JSON dictionary is changed by this function.
    """
    # get the original list of groups because we are going to change the
    # dict.keys() of json_dict inside the next list. Note: we have to convert
    # from dict_keys to list here otherwise the loop complains about changing
    # the dictionary during iteration
    buffer_names = list(json_dict.keys())
    for name in buffer_names:
        if name == "":
            raise ValueError("buffer name can't be empty")

        info = json_dict[name]  # changes to info will change json_dict[name]
        # make sure we have a key list
        if "key_list" not in info:
            raise ValueError(f"'{name}' is missing key_list")
            continue

        # find and expand any ranges in the key_list
        # do in a while loop with a controlled index since we are modifying
        # the length of the list inside the loop (be careful)
        i = 0
        while i < len(info["key_list"]):
            key_range = info["key_list"][i]
            # expand any 2-int lists
            if isinstance(key_range, list) and len(key_range) == 2:
                info["key_list"][i : i + 1] = range(key_range[0], key_range[1] + 1)
                i += key_range[1] - key_range[0]
            i += 1

        # Expand list_names if name contains a key-based formatter
        if "{key" in name:
            if (
                len(info["key_list"]) == 1
                and isinstance(info["key_list"][0], str)
                and "*" in info["key_list"][0]
            ):
                continue  # will be handled later, once the key_list is known
            for key in info["key_list"]:
                expanded_name = name.format(key=key)
                json_dict[expanded_name] = info.copy()
                json_dict[expanded_name]["key_list"] = [key]
            json_dict.pop(name)

    # now re-iterate and expand out_paths
    for name, info in json_dict.items():
        if len(info["key_list"]) == 1 and not (
            isinstance(info["key_list"][0], str) and "*" in info["key_list"][0]
        ):
            kw_dict["key"] = info["key_list"][0]
        if "out_stream" in info:
            if name != "*" and "{name" in info["out_stream"]:
                kw_dict["name"] = name
            try:
                info["out_stream"] = info["out_stream"].format(**kw_dict)
            except KeyError as msg:
                raise KeyError(
                    f"variable {msg} dereferenced in 'out_stream' not defined in kw_dict"
                )
            info["out_stream"] = os.path.expandvars(info["out_stream"])


def write_to_lh5_and_clear(
    raw_buffers: list[RawBuffer], lh5_store: LH5Store = None, **kwargs
) -> None:
    r"""Write a list of :class:`.RawBuffer`\ s to LH5 files and then clears them.

    Parameters
    ----------
    raw_buffers
        The list of ``RawBuffer``\ s to be written to file. Note: this is not a
        :class:`.RawBufferList` because the raw buffers may not have the
        same structure.  If a raw buffer has a `proc_spec` attribute, then
        :meth:`.buffer_processor.buffer_processor.buffer_processor` is used to
        process that buffer.
    lh5_store
        Allows user to send in a store holding a collection of already open
        files (saves some time opening / closing files).
    **kwargs
        keyword-arguments forwarded to
        :meth:`.lgdo.lh5_store.LH5Store.write_object`.

    See Also
    --------
    .lgdo.lh5_store.LH5Store.write_object
    """
    if lh5_store is None:
        lh5_store = lgdo.LH5Store()
    for rb in raw_buffers:
        if rb.lgdo is None or rb.loc == 0:
            continue  # no data to write
        ii = rb.out_stream.find(":")
        if ii == -1:
            filename = rb.out_stream
            group = "/"
        else:
            filename = rb.out_stream[:ii]
            group = rb.out_stream[ii + 1 :]
            if len(group) == 0:
                group = "/"  # in case out_stream ends with :

        # If a proc_spec if present for this RawBuffer, process that data and then write to the file!
        if rb.proc_spec is not None:
            # Perform the processing as requested in the `proc_spec` from `out_spec` in build_raw
            lgdo_to_write = buffer_processor(rb)
        else:
            lgdo_to_write = rb.lgdo

        # write if requested...
        if filename != "":
            lh5_store.write_object(
                lgdo_to_write,
                rb.out_name,
                filename,
                group=group,
                n_rows=rb.loc,
                **kwargs,
            )
        # and clear
        rb.loc = 0



"""
pygama convenience functions.
"""
import sys

import matplotlib.pyplot as plt
import numpy as np


def get_dataset_from_cmdline(args, run_db, cal_db):
    """
    make it easier to call this from argparse:
        arg("-ds", nargs='*', action="store", help="load runs for a DS")
        arg("-r", "--run", nargs=1, help="load a single run")
    """
    from pygama import DataSet

    if args["ds"]:
        ds_lo = int(args["ds"][0])
        try:
            ds_hi = int(args["ds"][1])
        except:
            ds_hi = None
        ds = DataSet(ds_lo, ds_hi, md=run_db, cal=cal_db, v=args["verbose"])

    if args["run"]:
        ds = DataSet(run=int(args["run"][0]), md=run_db, cal=cal_db,
                     v=args["verbose"])
    return ds


def sh(cmd, sh=False):
    """
    input a shell command as you would type it on the command line.
    """
    decoders = []
    for sub in DataTaker.__subclasses__():
        for subsub in sub.__subclasses__():
            try:
                decoder = subsub(object_info) # initialize the decoder
                decoders.append(decoder)
            except Exception as e:
                print(e)
                pass
    return decoders


def sizeof_fmt(num, suffix='B'):
    """
    given a file size in bytes, output a human-readable form.
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1024.0:
            return f"{num:.3f} {unit}{suffix}"
        num /= 1024.0
    return "{:.1f} {} {}".format(num, 'Y', suffix)


def set_plot_style(style):
    """
    Choose a pygama plot style.
    Current options: 'clint', 'root'
    Or add your own [label].mpl file in the pygama directory!
    """
    path = __file__.rstrip('.utils.py')
    plt.style.use(path+'/'+style+'.mpl')


def get_par_names(func):
    """
    Return a list containing the names of the arguments of "func" other than the
    first argument. In pygamaland, those are the function's "parameters."
    """
    from scipy._lib._util import getargspec_no_self
    args, varargs, varkw, defaults = getargspec_no_self(func)
    return args[1:]


def plot_func(func, pars, range=None, npx=None, **kwargs):
    """
    plot a function.  take care of the x-axis points automatically, or user can
    specify via range and npx arguments.
    """
    if npx is None:
        npx = 100
    if range is None:
        range = plt.xlim()
    xvals = np.linspace(range[0], range[1], npx)
    plt.plot(xvals, func(xvals, *pars), **kwargs)


def get_formatted_stats(mean, sigma, ndigs=2):
    """
    convenience function for formatting mean +/- sigma to the right number of
    significant figures.
    """
    if sigma == 0:
        fmt = '%d' % ndigs
        fmt = '%#.' + fmt + 'g'
        return fmt % mean, fmt % sigma
    sig_pos = int(np.floor(np.log10(abs(sigma))))
    sig_fmt = '%d' % ndigs
    sig_fmt = '%#.' + sig_fmt + 'g'
    mean_pos = int(np.floor(np.log10(abs(mean))))
    mdigs = mean_pos-sig_pos+ndigs
    if mdigs < ndigs-1: mdigs = ndigs - 1
    mean_fmt = '%d' % mdigs
    mean_fmt = '%#.' + mean_fmt + 'g'
    return mean_fmt % mean, sig_fmt % sigma


def print_fit_results(pars, cov, func=None, title=None, pad=True):
    """
    convenience function for scipy.optimize.curve_fit results
    """
    if title is not None:
        print(title+":")
    if func is None:
        for i in range(len(pars)): par_names.append("p"+str(i))
    else:
        par_names = get_par_names(func)
    for i in range(len(pars)):
        mean, sigma = get_formatted_stats(pars[i], np.sqrt(cov[i][i]))
        print(par_names[i], "=", mean, "+/-", sigma)
    if pad:
        print("")


def tree_draw(tree, vars, tcut):
    """
    if you have to debase yourself and use ROOT, this is an easy
    convenience function for quickly extracting data from TTrees.
    TTree::Draw can only handle groups of 4 variables at a time,
    but here we can put in as many as we want, and
    return a list of numpy.ndarrays for each one
    """
    var_list = vars.split(":")
    np_arrs = [[] for v in var_list]

    for i in range(0, len(var_list), 4):

        tmp_list = var_list[i:i + 4]
        tmp_draw = ":".join(tmp_list)
        n = tree.Draw(tmp_draw, tcut, "goff")

        for j, var in enumerate(tmp_list):
            # print(i, j, var, "getting V", j+1, "writing to np_arrs", i + j)
            tmp = getattr(tree, f"GetV{j + 1}")()
            np_arrs[i + j] = np.array([tmp[k] for k in range(n)])

    return tuple(np_arrs)


def peakdet(v, delta, x=None):
    """
    Converted from MATLAB script at: http://billauer.co.il/peakdet.html
    Returns two arrays: [maxtab, mintab] = peakdet(v, delta, x)
    An updated (vectorized) version is in pygama.dsp.transforms.peakdet
    """
    maxtab, mintab = [], []

    if x is None:
        x = np.arange(len(v))
    v = np.asarray(v)

    # sanity checks
    if len(v) != len(x):
        exit("Input vectors v and x must have same length")
    if not np.isscalar(delta):
        exit("Input argument delta must be a scalar")
    if delta <= 0:
        exit("Input argument delta must be positive")

    maxes, mins = [], []
    min, max = np.inf, -np.inf
    find_max = True
    for i in range(len(x)):

        # for i=0, all 4 of these get set
        if v[i] > max:
            max, imax = v[i], x[i]
        if v[i] < min:
            min, imin = v[i], x[i]

        if find_max:
            # if the sample is less than the current max,
            # declare the previous one a maximum, then set this as the new "min"
            if v[i] < max - delta:
                maxes.append((imax, max))
                min, imin = v[i], x[i]
                find_max = False
        else:
            # if the sample is more than the current min,
            # declare the previous one a minimum, then set this as the new "max"
            if v[i] > min + delta:
                mins.append((imin, min))
                max, imax = v[i], x[i]
                find_max = True

    return np.array(maxes), np.array(mins)


def linear_fit_by_sums(x, y, var=1):
    """
    Fast computation of weighted linear least squares fit to a linear model

    Note: doesn't compute covariances. If you want covariances, just use polyfit

    Parameters
    ----------
    x : array like
        x values for the fit
    y : array like
        y values for the fit
    var : array like (optional)
        The variances for each y-value

    Returns
    -------
    (m, b) : tuple (float, float)
        The slope (m) and y-intercept (b) of the best fit (in the least-squares
        sense) of the data to y = mx + b
    """
    y = y/var
    x = x/var
    sum_wts = len(y)/var if np.isscalar(var) else np.sum(1/var)
    sum_x = np.sum(x)
    sum_xx = np.sum(x*x)
    sum_y = np.sum(y)
    sum_yx = np.sum(y*x)
    m = (sum_wts * sum_yx - sum_y * sum_x) / (sum_wts * sum_xx - sum_x**2)
    b = (sum_y - m * sum_x) / sum_wts
    return m, b


def fit_simple_scaling(x, y, var=1):
    """
    Fast computation of weighted linear least squares fit to a simple scaling

    I.e. y = scale * x. Returns the best fit scale parameter and its variance.

    Parameters
    ----------
    x : array like
        x values for the fit
    y : array like
        y values for the fit
    var : array like (optional)
        The variances for each y-value

    Returns
    -------
    scale, scale_var: tuple (float, float)
        The scale parameter and its variance
    """
    x = np.asarray(x)
    y = np.asarray(y)
    scale_var = 1/np.sum(x*x/var)
    scale = np.sum(y*x/var) * scale_var
    return scale, scale_var



from __future__ import annotations

import numpy as np
from numba import guvectorize

from pygama.dsp.errors import DSPFatal
from pygama.dsp.utils import numba_defaults_kwargs as nb_kwargs


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:])",
        "void(float64[:], float64[:], float64[:])",
    ],
    "(n),(m),(p)",
    **nb_kwargs,
)
def histogram(
    w_in: np.ndarray, weights_out: np.ndarray, borders_out: np.ndarray
) -> None:
    """Produces and returns an histogram of the waveform.

    Parameters
    ----------
    w_in
        Data to be histogrammed.
    weights_out
        The output histogram weights.
    borders_out
        The output histogram bin edges of the histogram. Length must be len(weights_out)+1

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "hist_weights, hist_borders": {
            "function": "histogram",
            "module": "pygama.dsp.processors.histogram",
            "args": ["waveform", "hist_weights(100)", "hist_borders(101)"],
            "unit": ["none", "ADC"]
        }

    Note
    ----
    This implementation is significantly faster than just wrapping
    :func:`numpy.histogram`.

    See Also
    --------
    .histogram_stats
    """

    if len(weights_out) + 1 != len(borders_out):
        raise DSPFatal("length borders_out must be exactly 1 + length of weights_out")

    weights_out[:] = 0
    borders_out[:] = np.nan

    if np.isnan(w_in).any():
        return

    # create the bin borders
    borders_out[0] = min(w_in)
    delta = 0

    # number of bins
    bin_in = len(weights_out)

    # define the bin edges
    delta = (max(w_in) - min(w_in)) / (bin_in)
    for i in range(0, bin_in, 1):
        borders_out[i + 1] = min(w_in) + delta * (i + 1)

    # make the histogram
    for i in range(0, len(w_in), 1):
        for k in range(1, len(borders_out), 1):
            if (w_in[i] - borders_out[k]) < 0:
                weights_out[k - 1] += 1
                break


@guvectorize(
    [
        "void(float32[:], float32[:], float32[:], float32[:],float32[:],float32)",
        "void(float64[:], float64[:], float64[:], float64[:],float64[:],float64)",
    ],
    "(n),(m),(),(),(),()",
    **nb_kwargs,
)
def histogram_stats(
    weights_in: np.ndarray,
    edges_in: np.ndarray,
    mode_out: int,
    max_out: float,
    fwhm_out: float,
    max_in: float,
) -> None:
    """Compute useful histogram-related quantities.

    Parameters
    ----------
    weights_in
        histogram weights.
    edges_in
        histogram bin edges.
    max_in
        if not :any:`numpy.nan`, this value is used as the histogram bin
        content at the mode.  Otherwise the mode is computed automatically
        (from left to right).
    mode_out
        the computed mode of the histogram. If `max_in` is not :any:`numpy.nan`
        then the closest waveform index to `max_in` is returned.
    max_out
        the histogram bin content at the mode.
    fwhm_out
        the FWHM of the histogram, calculated by starting from the mode and
        descending left and right.

    JSON Configuration Example
    --------------------------

    .. code-block :: json

        "fwhm, idx_out, max_out": {
            "function": "histogram_stats",
            "module": "pygama.dsp.processors.histogram",
            "args": ["hist_weights","hist_borders","idx_out","max_out","fwhm","np.nan"],
            "unit": ["ADC", "none", "ADC"]
        }

    See Also
    --------
    .histogram
    """

    fwhm_out[0] = np.nan
    mode_out[0] = np.nan
    max_out[0] = np.nan

    if np.isnan(weights_in).any():
        return

    if len(weights_in) + 1 != len(edges_in):
        raise DSPFatal("length edges_in must be exactly 1 + length of weights_in")

    # find global maximum search from left to right
    max_index = 0
    if np.isnan(max_in):
        for i in range(0, len(weights_in), 1):
            if weights_in[i] > weights_in[max_index]:
                max_index = i

    # is user specifies mean justfind mean index
    else:
        if max_in > edges_in[-2]:
            max_index = len(weights_in) - 1
        else:
            for i in range(0, len(weights_in), 1):
                if abs(max_in - edges_in[i]) < abs(max_in - edges_in[max_index]):
                    max_index = i

    mode_out[0] = max_index
    # returns left bin edge
    max_out[0] = edges_in[max_index]

    # and the approx fwhm
    for i in range(max_index, len(weights_in), 1):
        if weights_in[i] <= 0.5 * weights_in[max_index] and weights_in[i] != 0:
            fwhm_out[0] = abs(max_out[0] - edges_in[i])
            break

    # look also into the other direction
    for i in range(0, max_index, 1):
        if weights_in[i] >= 0.5 * weights_in[max_index] and weights_in[i] != 0:
            if fwhm_out[0] < abs(max_out[0] - edges_in[i]):
                fwhm_out[0] = abs(max_out[0] - edges_in[i])
            break
